"""Utility functions for interacting with the Relay service"""
from __future__ import annotations

import asyncio
import base64
import datetime
import functools
import inspect
import json
import signal
import weakref
from typing import (Any, Awaitable, Callable, Iterable, Mapping, Optional,
                    Protocol, Union)


def json_object_hook(dct: Mapping[str, Any]) -> Any:
    if '$encoding' in dct:
        try:
            decoder: Callable[[str], str] = {
                'base64': base64.standard_b64decode,
                '': lambda data: data,
            }[dct['$encoding']]

            return decoder(dct['data'])
        except KeyError:
            # Either dct does not contain data or has an encoding that we can't
            # handle.
            pass
    return dct


class JSONEncoder(json.JSONEncoder):

    @functools.singledispatchmethod
    def default(self, obj: Any) -> Any:
        try:
            it = iter(obj)
        except TypeError:
            pass
        else:
            return list(it)

        return super(JSONEncoder, self).default(obj)

    @default.register
    def _datetime(self, obj: datetime.datetime) -> str:
        return obj.isoformat()

    @default.register
    def _bytes(self, obj: bytes) -> Union[str, Mapping[str, Any]]:
        try:
            return obj.decode('utf-8')
        except UnicodeDecodeError:
            return {
                '$encoding': 'base64',
                'data': base64.standard_b64encode(obj),
            }


def is_async_callable(obj: Any) -> bool:
    if not callable(obj):
        return False

    return (
        inspect.iscoroutinefunction(obj) or
        inspect.iscoroutinefunction(obj.__call__)
    )


TerminationEvent = Callable[[], Awaitable[None]]


class TerminationPolicy(Protocol):

    async def attach(self) -> Optional[TerminationEvent]: ...


class NoTerminationPolicy(TerminationPolicy):

    async def attach(self) -> Optional[TerminationEvent]:
        return None


class SoftTerminationPolicy(TerminationPolicy):

    _tasks: weakref.WeakKeyDictionary[asyncio.Task[Any], asyncio.Event]
    _timeout_sec: Optional[float]

    def __init__(self, *, timeout_sec: Optional[float] = None):
        self._tasks = weakref.WeakKeyDictionary()
        self._timeout_sec = timeout_sec

    async def _terminate_task(self, task: asyncio.Task[Any]) -> None:
        event = self._tasks.get(task)
        if event is not None:
            event.set()

        if task.done():
            return

        if self._timeout_sec is not None:
            loop = asyncio.get_running_loop()
            loop.call_later(self._timeout_sec, task.cancel)

    def terminate_task(self, task: asyncio.Task[Any]) -> None:
        asyncio.run_coroutine_threadsafe(
            self._terminate_task(task),
            task.get_loop(),
        )

    def terminate_all(self) -> None:
        for task in self._tasks:
            self.terminate_task(task)

    async def attach(self) -> Optional[TerminationEvent]:
        task = asyncio.current_task()
        assert task is not None

        try:
            event = self._tasks[task]
        except KeyError:
            event = asyncio.Event()
            self._tasks[task] = event

        async def wait() -> None:
            await event.wait()

        return wait


class SignalTerminationPolicy(TerminationPolicy):

    _signals: Iterable[signal.Signals]
    _delegate: SoftTerminationPolicy

    def __init__(self, *,
                 signals: Optional[Iterable[signal.Signals]] = None,
                 timeout_sec: Optional[float] = None):
        if signals is None:
            signals = [signal.SIGINT, signal.SIGTERM]

        self._signals = signals
        self._delegate = SoftTerminationPolicy(timeout_sec=timeout_sec)

    async def attach(self) -> Optional[TerminationEvent]:
        loop = asyncio.get_running_loop()
        task = asyncio.current_task()
        assert task is not None

        event = self._delegate.attach()

        for sig in self._signals:
            loop.add_signal_handler(sig, self._delegate.terminate_task, task)

        return await event



from requests_mock import Adapter

from relay_sdk.client import new_session
from relay_sdk.outputs import Outputs


class TestOutputs:

    def test_set(self, requests_mock: Adapter) -> None:
        requests_mock.register_uri(
            'PUT', 'http+api://api/outputs/foo',
            text='OK',
            request_headers={'content-type': 'application/json'},
            additional_matcher=lambda request: request.text == r'"bar"',
        )
        Outputs(new_session(api_url='http://api')).set('foo', 'bar')



from requests_mock import Adapter

from relay_sdk.client import new_session
from relay_sdk.events import Events


class TestEvents:

    def test_emit(self, requests_mock: Adapter) -> None:
        requests_mock.register_uri(
            'POST', 'http+api://api/events',
            text='OK',
            request_headers={'content-type': 'application/json'},
            additional_matcher=lambda request:
                request.json() == {'data': {'foo': 'bar'}},
        )
        Events(new_session(api_url='http://api')).emit({'foo': 'bar'})

    def test_emit_with_key(self, requests_mock: Adapter) -> None:
        requests_mock.register_uri(
            'POST', 'http+api://api/events',
            text='OK',
            request_headers={'content-type': 'application/json'},
            additional_matcher=lambda request:
                request.json() == {'data': {'foo': 'bar'}, 'key': 'key'},
        )
        Events(new_session(api_url='http://api')
               ).emit({'foo': 'bar'}, key='key')



import pytest
from requests_mock import Adapter

from relay_sdk.interface import Dynamic, Interface, UnresolvableException


@pytest.mark.parametrize(
    'test_input, expected',
    [
        (Dynamic.foo.bar, 'foo["bar"]'),
        (Dynamic.foo, 'foo'),
        (Dynamic.very.deep.nesting, 'very["deep"]["nesting"]'),
    ],
)
def test_dynamic(test_input: Dynamic, expected: str) -> bool:
    return str(test_input) == expected


class TestInterface:

    def test_all(self, requests_mock: Adapter) -> None:
        requests_mock.register_uri(
            'GET', 'http+api://api/spec',
            text=r'{"value": {"foo": "bar"}, "complete": true}',
        )
        assert Interface(api_url='http://api').get() == {'foo': 'bar'}

    def test_query(self, requests_mock: Adapter) -> None:
        requests_mock.register_uri(
            'GET', 'http+api://api/spec?q=foo',
            complete_qs=True,
            text=r'{"value": "bar", "complete": true}',
        )
        assert Interface(api_url='http://api').get(Dynamic.foo) \
            == 'bar'

    def test_incomplete(self, requests_mock: Adapter) -> None:
        requests_mock.register_uri(
            'GET', 'http+api://api/spec',
            text=r'''{
                "value": {"foo": {"$type": "Secret", "name": "foo"}},
                "complete": false}
            ''',
        )
        with pytest.raises(UnresolvableException):
            Interface(api_url='http://api').get()



"""Local data types"""
from typing import Text, Tuple, Union

HTTPTimeout = Union[float, Tuple[float, float], Tuple[float, None]]
HTTPClientCertificate = Union[
    bytes,
    Text,
    Tuple[Union[bytes, Text], Union[bytes, Text]],
]



"Outputs allow a step to write data that will be available for later steps"
import json
import logging
from typing import Any
from urllib.parse import quote

from requests import Session

from .util import JSONEncoder

logger = logging.getLogger(__name__)


class Outputs:
    """Use Outputs to write keys to the metadata service
        that later steps can retrieve"""
    def __init__(self, client: Session) -> None:
        self._client = client

    def set(self, name: str, value: Any) -> None:
        """Set writes a value of a given name to the service

        Args:
            name: a string containing the key name to write
            value: the value of the key to be written. Can
                be any data type that can serialize to JSON
        """
        r = self._client.put(
            'http+api://api/outputs/{0}'.format(quote(name)),
            data=json.dumps(value, cls=JSONEncoder),
            headers={'content-type': 'application/json'},
        )
        r.raise_for_status()

        logger.info('Set output %s', repr(name))



import asyncio
import datetime
import json
import signal
from typing import Any

import pytest

from relay_sdk.util import (JSONEncoder, SignalTerminationPolicy,
                            is_async_callable, json_object_hook)


@pytest.mark.parametrize(
    'test_input, expected',
    [
        (
            {'foo': b'\x90'},
            r'{"foo": {"$encoding": "base64", "data": "kA=="}}',
        ),
        (range(0, 5), r'[0, 1, 2, 3, 4]'),
        (
            datetime.datetime(2020, 1, 2, 3, 4, 5, 0, datetime.timezone.utc),
            r'"2020-01-02T03:04:05+00:00"',
        ),
    ],
)
def test_json_encoding(test_input: Any, expected: Any) -> None:
    assert json.dumps(test_input, cls=JSONEncoder, sort_keys=True) == expected


def test_json_decoding() -> None:
    assert json.loads(
        r'{"$encoding": "base64", "data": "kA=="}',
        object_hook=json_object_hook,
    ) == b'\x90'


def test_is_async_callable() -> None:
    class AsyncCallable:

        async def __call__(self) -> None:
            pass

    class Callable:

        def __call__(self) -> None:
            pass

    async def async_callable() -> None:
        pass

    def sync_callable() -> None:
        pass

    assert is_async_callable(AsyncCallable())
    assert is_async_callable(async_callable)
    assert not is_async_callable(Callable)
    assert not is_async_callable(sync_callable)
    assert not is_async_callable(2)


class TestSignalTerminationPolicy:

    @pytest.mark.asyncio
    async def test_indefinite(self) -> None:
        pol = SignalTerminationPolicy(signals=[signal.SIGWINCH])

        ready_event = asyncio.Event()
        completion_event = asyncio.Event()

        async def run() -> None:
            waiter = await pol.attach()
            assert waiter is not None
            ready_event.set()

            await waiter()
            completion_event.set()

        async def raise_signal() -> None:
            await ready_event.wait()
            signal.raise_signal(signal.SIGWINCH)

        await asyncio.gather(run(), raise_signal())
        await asyncio.wait_for(completion_event.wait(), 60)

    @pytest.mark.asyncio
    async def test_timeout(self) -> None:
        pol = SignalTerminationPolicy(
            signals=[signal.SIGWINCH],
            timeout_sec=0,  # Immediate timeout.
        )

        ready_event = asyncio.Event()

        async def run() -> None:
            waiter = await pol.attach()
            assert waiter is not None
            ready_event.set()

            await waiter()

            # Instead of completing we'll force the policy to terminate us
            # early.
            await asyncio.sleep(30)

        async def raise_signal() -> None:
            await ready_event.wait()
            signal.raise_signal(signal.SIGWINCH)

        with pytest.raises(asyncio.CancelledError):
            await asyncio.gather(run(), raise_signal())



"Generating events for the service to act on"
import json
from typing import Any, Dict, Mapping, Optional

from requests import Session

from .util import JSONEncoder


class Events:
    """Class for generating event payloads to the service API"""

    def __init__(self, client: Session) -> None:
        self._client = client

    def emit(self, data: Mapping[str, Any], key: Optional[str] = None) -> None:
        """Sends an event to the service.

        Use this from a Trigger handler to start the workflow
        associated with the trigger.

        Accepts a Mapping of data fields from the request payload
        that came in to the Trigger handler.

        A key can be optionally provided to uniquely identify
        the event."""

        post_data: Dict[str, Any] = {'data': data}
        if key:
            post_data['key'] = key

        r = self._client.post(
            'http+api://api/events',
            data=json.dumps(post_data, cls=JSONEncoder),
            headers={'content-type': 'application/json'},
        )

        r.raise_for_status()



from os import path

import setuptools

long_description_file = path.join(
    path.abspath(path.dirname(__file__)),
    'README.rst',
)
with open(long_description_file, encoding='utf-8') as fp:
    long_description = fp.read()

setuptools.setup(
    name='relay-sdk',
    use_scm_version={
        'relative_to': __file__,
    },
    author='Puppet, Inc.',
    author_email='relay@puppet.com',
    description='SDK for interacting with Puppet Relay',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/puppetlabs/relay-sdk-python',
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.8',
    setup_requires=[
        'setuptools_scm',
    ],
    install_requires=[
        'asgiref>=3.2.7',
        'hypercorn>=0.13.2',
        'requests>=2.23',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
    ],
)
